from typing import Callable, Optional, Any, Tuple, TypeVar
from urllib.parse import urlparse

from pika import ConnectionParameters, PlainCredentials, BlockingConnection, BasicProperties, spec
from pika.adapters.blocking_connection import BlockingChannel

from unipipeline.modules.uni_broker import UniBroker, UniBrokerMessageManager
from unipipeline.modules.uni_broker_definition import UniBrokerDefinition, UniMessageCodec
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_meta import UniMessageMeta
from unipipeline.utils.connection_pool import ConnectionObj, TConnectionObj, connection_pool

BASIC_PROPERTIES__HEADER__COMPRESSION_KEY = 'compression'


TMessage = TypeVar('TMessage', bound=UniMessage)


class AmqpUniBrokerMessageManager(UniBrokerMessageManager):

    def __init__(self, channel: BlockingChannel, method_frame: spec.Basic.Deliver) -> None:
        self._channel = channel
        self._method_frame = method_frame
        self._acked = False

    def reject(self) -> None:
        self._channel.basic_reject(delivery_tag=self._method_frame.delivery_tag, requeue=True)

    def ack(self) -> None:
        if self._acked:
            return
        self._acked = True
        self._channel.basic_ack(delivery_tag=self._method_frame.delivery_tag)


class RMQConnectionObj(ConnectionObj[BlockingConnection]):

    def __init__(self, params: ConnectionParameters) -> None:
        self._params = params
        self._connection: Optional[BlockingConnection] = None

    def __hash__(self) -> int:
        return hash(f'{self._params.host}{self._params.port}{self._params.credentials.username}{self._params.credentials.password}')

    def get(self) -> TConnectionObj:
        assert self._connection is not None
        return self._connection

    def is_closed(self) -> bool:
        return self._connection is None or self._connection.is_closed

    def connect(self) -> None:
        assert self._connection is None
        self._connection = BlockingConnection(self._params)

    def close(self):
        if self._connection is not None:
            if not self._connection.is_closed:
                self._connection.close()
            self._connection = None


class AmqpUniBroker(UniBroker):
    @classmethod
    def get_connection_uri(cls) -> str:
        raise NotImplementedError(f"cls method get_connection_uri must be implemented for class '{cls.__name__}'")

    def get_exchange_name(self) -> str:
        return self.definition.rmq_definition.exchange_name

    def __init__(self, definition: UniBrokerDefinition[bytes]) -> None:
        super().__init__(definition)

        self._connection_key = self.get_connection_uri()
        url_params_pr = urlparse(url=self._connection_key)

        params = ConnectionParameters(
            heartbeat=self.definition.rmq_definition.heartbeat,
            blocked_connection_timeout=self.definition.rmq_definition.blocked_connection_timeout,
            socket_timeout=self.definition.rmq_definition.socket_timeout,
            stack_timeout=self.definition.rmq_definition.stack_timeout,
            retry_delay=self.definition.retry_delay_s,
            host=url_params_pr.hostname,
            port=url_params_pr.port,
            credentials=PlainCredentials(url_params_pr.username, url_params_pr.password, erase_on_connect=False),
        )

        self._connector = connection_pool.new_manager(RMQConnectionObj(params))

        self._read_channel: Optional[BlockingChannel] = None
        self._write_channel: Optional[BlockingChannel] = None
        self._consuming_started = False

    def connect(self) -> None:
        self._connector.connect()

    def close(self) -> None:
        self._connector.close()

    def _bind(self, is_read: bool, topic: str) -> None:
        if is_read:
            channel = self._read_channel
        else:
            channel = self._write_channel

        if channel is not None:
            return

        channel = self._connector.connect().channel()
        exchange = self.get_exchange_name()
        channel.exchange_declare(
            exchange=exchange,
            exchange_type=self.definition.rmq_definition.exchange_type,
            passive=self.definition.passive,
            durable=self.definition.durable,
            auto_delete=self.definition.auto_delete,
        )

        channel.queue_declare(
            queue=topic,
            durable=self.definition.durable,
            auto_delete=self.definition.auto_delete
        )
        channel.queue_bind(queue=topic, exchange=exchange, routing_key=topic)

        if is_read:
            self._read_channel = channel
        else:
            self._write_channel = channel

    def consume(
        self,
        topic: str,
        processor: Callable[[UniMessageMeta, UniBrokerMessageManager], None],
        consumer_tag: str,
        worker_name: str,
        prefetch: int = 1
    ) -> None:
        if self._consuming_started:
            raise ConnectionError('you cannot consume twice!')
        self._consuming_started = True
        self._bind(True, topic)
        assert self._read_channel is not None
        self._read_channel.basic_qos(prefetch_count=prefetch)
        self._read_channel.basic_consume(
            queue=topic,
            on_message_callback=self._wrap_message(processor),
            consumer_tag=consumer_tag,
        )
        self._read_channel.start_consuming()

    def serialize_body(self, meta: UniMessageMeta) -> Tuple[bytes, BasicProperties]:
        meta_dumps = self.definition.message_codec.dumps(meta.dict())
        meta_compressed = self.definition.message_codec.compress(meta_dumps)

        properties = BasicProperties(
            content_type=self.definition.message_codec.content_type,
            content_encoding='utf-8',
            delivery_mode=2 if self.definition.is_persistent else 0,
            headers={BASIC_PROPERTIES__HEADER__COMPRESSION_KEY: self.definition.message_codec.compression}
        )
        return meta_compressed, properties

    def parse_body(self, body: bytes, properties: BasicProperties) -> UniMessageMeta:
        codec: UniMessageCodec[Any] = UniMessageCodec(
            compression=properties.headers.get(BASIC_PROPERTIES__HEADER__COMPRESSION_KEY, None),
            encoding=properties.content_encoding,
            content_type=properties.content_type
        )
        body_uncompressed = codec.decompress(body)
        body_json = codec.loads(body_uncompressed)
        return UniMessageMeta(**body_json)

    def _wrap_message(self, processor: Callable[[UniMessageMeta, UniBrokerMessageManager], None]) -> Callable[[BlockingChannel, Any, Any, bytes], None]:
        def wrapper(channel: BlockingChannel, method_frame: spec.Basic.Deliver, properties: BasicProperties, body: bytes) -> None:
            manager = AmqpUniBrokerMessageManager(channel, method_frame)
            meta = self.parse_body(body, properties)
            processor(meta, manager)

        return wrapper

    def publish(self, topic: str, meta: UniMessageMeta) -> None:
        self._bind(False, topic)

        body, properties = self.serialize_body(meta)

        assert self._write_channel is not None
        self._write_channel.basic_publish(
            exchange=self.get_exchange_name(),
            routing_key=topic,
            body=body,
            properties=properties
        )



from typing import Dict, Type, TypeVar, Any

from unipipeline.modules.uni_broker import UniBroker
from unipipeline.modules.uni_broker_definition import UniBrokerDefinition
from unipipeline.modules.uni_message_type_definition import UniMessageTypeDefinition
from unipipeline.modules.uni_waiting_definition import UniWaitingDefinition
from unipipeline.modules.uni_worker_definition import UniWorkerDefinition

TMessage = TypeVar('TMessage')
T = TypeVar('T')


class UniDefinition:
    def __init__(
        self,
        brokers: Dict[str, UniBrokerDefinition[Any]],
        workers: Dict[str, UniWorkerDefinition],
        waitings: Dict[str, UniWaitingDefinition],
        messages: Dict[str, UniMessageTypeDefinition],
    ) -> None:
        self._brokers = brokers
        self._workers = workers
        self._waitings = waitings
        self._messages = messages
        self._connected_brokers: Dict[str, UniBroker] = dict()
        self._worker_definition_by_type: Dict[Any, UniWorkerDefinition] = dict()

    def get_worker(self, name: str) -> UniWorkerDefinition:
        return self._workers[name]

    @property
    def brokers(self) -> Dict[str, UniBrokerDefinition[Any]]:
        return self._brokers

    @property
    def workers(self) -> Dict[str, UniWorkerDefinition]:
        return self._workers

    @property
    def waitings(self) -> Dict[str, UniWaitingDefinition]:
        return self._waitings

    @property
    def messages(self) -> Dict[str, UniMessageTypeDefinition]:
        return self._messages

    def get_connected_broker_instance(self, name: str) -> UniBroker:
        if name not in self._connected_brokers:
            self._connected_brokers[name] = UniBroker.waiting_for_connection(self._brokers[name])
        return self._connected_brokers[name]

    def wait_related_brokers(self, worker_name: str) -> None:
        broker_names = self._workers[worker_name].get_related_broker_names(self._workers)
        for bn in broker_names:
            self.get_connected_broker_instance(bn)

    def load_workers(self, uni_type: Type[T]) -> None:
        if self._worker_definition_by_type:
            return
        for wd in self._workers.values():
            self._worker_definition_by_type[wd.type.import_class(uni_type)] = wd

    def get_worker_definition_by_type(self, worker_type: Any, uni_type: Type[T]) -> UniWorkerDefinition:
        self.load_workers(uni_type)
        return self._worker_definition_by_type[worker_type]



import uuid
from datetime import datetime
from typing import Dict, Any, Optional
from uuid import UUID

from pydantic import BaseModel


class UniMessageMeta(BaseModel):
    id: UUID
    date_created: datetime
    payload: Dict[str, Any]

    parent: Optional[Dict[str, Any]]

    @staticmethod
    def create_new(data: Dict[str, Any]) -> 'UniMessageMeta':
        return UniMessageMeta(
            id=uuid.uuid4(),
            date_created=datetime.now(),
            payload=data,
            parent=None
        )

    def create_child(self, data: Dict[str, Any]) -> 'UniMessageMeta':
        return UniMessageMeta(
            id=uuid.uuid4(),
            date_created=datetime.now(),
            payload=data,
            parent=self.dict()
        )



from typing import Dict, List, Any

import yaml

from unipipeline.modules.uni_broker import UniBroker
from unipipeline.modules.uni_broker_definition import UniBrokerDefinition, UniMessageCodec, UniBrokerRMQPropsDefinition, UniBrokerKafkaPropsDefinition
from unipipeline.modules.uni_definition import UniDefinition
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_type_definition import UniMessageTypeDefinition
from unipipeline.modules.uni_waiting_definition import UniWaitingDefinition
from unipipeline.modules.uni_wating import UniWaiting
from unipipeline.modules.uni_worker import UniWorker
from unipipeline.modules.uni_worker_definition import UniWorkerDefinition, WORKER_ERROR_NAME
from unipipeline.utils.parse_definition import parse_definition
from unipipeline.utils.parse_type import parse_type
from unipipeline.utils.template import template


class Uni:
    def __init__(self, config_file_path: str) -> None:
        self._config_file_path = config_file_path

        with open(config_file_path, "rt") as f:
            self._config = yaml.safe_load(f)

        self._definition = UniDefinition(
            waitings=self._parse_waitings(),
            brokers=self._parse_brokers(),
            messages=self._parse_messages(),
            workers=self._parse_workers(),
        )

    def check_load_all(self) -> None:
        for b in self._definition.brokers.values():
            b.type.import_class(UniBroker)

        for m in self._definition.messages.values():
            m.type.import_class(UniMessage)

        for worker in self._definition.workers.values():
            worker.type.import_class(UniWorker)

        for waiting in self._definition.waitings.values():
            waiting.type.import_class(UniWaiting)

    def get_worker(self, name: str) -> UniWorker[Any]:
        definition = self._definition.get_worker(name)
        worker_type = definition.type.import_class(UniWorker)
        w = worker_type(definition=definition, index=self._definition)
        return w

    def _parse_messages(self) -> Dict[str, UniMessageTypeDefinition]:
        result = dict()
        for definition in parse_definition(self._config["messages"]):
            name = definition["name"]
            data = dict(
                name=name,
                version=definition["version"],
            )
            result[name] = UniMessageTypeDefinition(
                **data,
                type=parse_type(template(definition["type"], data)),
            )
        return result

    def _parse_waitings(self) -> Dict[str, UniWaitingDefinition]:
        result = dict()
        for definition in parse_definition(self._config['waitings']):
            name = definition["name"]
            data = dict(
                name=name,
                retry_max_count=definition["retry_max_count"],
                retry_delay_s=definition["retry_delay_s"],
            )
            result[name] = UniWaitingDefinition(
                **data,
                type=parse_type(template(definition["type"], data)),
            )
        return result

    def _parse_brokers(self) -> Dict[str, UniBrokerDefinition[Any]]:
        result: Dict[str, Any] = dict()
        for definition in parse_definition(self._config["brokers"]):
            name = definition["name"]
            data = dict(
                name=name,
                retry_max_count=definition["retry_max_count"],
                retry_delay_s=definition["retry_delay_s"],
                passive=definition["passive"],
                durable=definition["durable"],
                auto_delete=definition["auto_delete"],
                is_persistent=definition["is_persistent"],
            )

            result[name] = UniBrokerDefinition(
                **data,
                type=parse_type(template(definition["type"], data)),
                message_codec=UniMessageCodec(
                    content_type=definition["content_type"],
                    compression_content_type=definition["compression"],
                    encoding=definition["encoding"],
                ),
                rmq_definition=UniBrokerRMQPropsDefinition(
                    exchange_name=definition['exchange_name'],
                    heartbeat=definition['heartbeat'],
                    blocked_connection_timeout=definition['blocked_connection_timeout'],
                    socket_timeout=definition['socket_timeout'],
                    stack_timeout=definition['stack_timeout'],
                    exchange_type=definition['exchange_type'],
                ),
                kafka_definition=UniBrokerKafkaPropsDefinition(
                    bootstrap_servers=definition['bootstrap_servers'],
                    api_version=definition['api_version'],
                )

            )
        return result

    def _parse_workers(self) -> Dict[str, UniWorkerDefinition]:
        result = dict()

        out_workers = set()

        for definition in parse_definition(self._config["workers"], {WORKER_ERROR_NAME, }):
            name = definition["name"]

            assert isinstance(definition["output_workers"], list)
            output_workers = definition["output_workers"]

            for ow in definition["output_workers"]:
                assert isinstance(ow, str), f"ow must be str. {type(ow)} given"
                out_workers.add(ow)

            broker = self._parse_brokers()[definition["broker"]]
            input_message = self._parse_messages()[definition["input_message"]]

            watings: List[UniWaitingDefinition] = list()
            for w in definition["waiting_for"]:
                watings.append(self._parse_waitings()[w])

            data = dict(
                name=name,
                broker=broker,
                is_permanent=definition["is_permanent"],
                auto_ack=definition["auto_ack"],
                prefetch=definition["prefetch"],
                input_message=input_message,
                retry_max_count=definition["retry_max_count"],
                retry_delay_s=definition["retry_delay_s"],
                max_ttl_s=definition["max_ttl_s"],
            )

            defn = UniWorkerDefinition(
                type=parse_type(template(definition["type"], data)),
                topic=template(definition["topic"], data),
                output_workers=output_workers,
                waitings=watings,
                **data,
            )

            result[name] = defn

        assert set(result.keys()).intersection(out_workers), f"invalid workers relations: {out_workers.difference(set(result.keys()))}"

        return result
